# different work location
locations = {
    "rvk-office": {"desc": "Reykjavík office"},
    "hel-office": {"desc": "Helsinki office"},
    "tailscale": {"desc": "working over tailscale network, e.g. from home"},
    "external": {"desc": "working from the internet without VPN"},
}
